/***************************************************************************
 *   Copyright (C) 2005 by Oliver Saal                                     *
 *   http://www.oliver-saal.de/                                            *
 *   osaal@gmx.de                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/ 


#include "data.h"

#include <qobject.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qstringlist.h>
#include <qregexp.h>
#include <qapplication.h>
#include <qmessagebox.h>
#include <qprogressdialog.h>

double GetCostsPerKm (const unsigned int uPollutant, const unsigned int uAxes)
{
unsigned int uRet=0;
	if (uPollutant > 2 || uAxes > 1) return 0;
	return (0.09 + (double)uPollutant*0.02 + (double)uAxes/100.0);
}

CNode::CNode ()
{
	m_iBAB = 0;
	m_status = StatusOff;
	m_typ = TypUnknown;
	m_pNearestLink = NULL;
	m_fNearestKm = 0.0;

	m_listLinkOut.setAutoDelete (false);
	m_listLinkIn.setAutoDelete (false);

	m_uCSVLine = 0;
	m_pCanvasNode = NULL;
}

CNode::~CNode ()
{
}

CNode::typ CNode::GetTyp (const QString strShortcut)
{
QString str = strShortcut.stripWhiteSpace().upper();
	if (str == "AD")
		return TypAD;
	else if (str == "AK")
		return TypAK;
	else if (str == "AS")
		return TypAS;
	else if (str == "AN")
		return TypAN;
	else if (str == "BS")
		return TypBS;
	else if (str == "BG")
		return TypBG;
	else if (str == "TR")
		return TypTR;
	else
		return TypUnknown;
}

CNode::status CNode::GetStatus (const char c)
{
	switch (c)
	{
	case 'r':
	case '1': return StatusOn;
	case '0': return StatusOff;
	case 'g': return StatusUnderConstruction;
	default: return StatusOff;
	}
}

QString CNode::GetStatusText () const
{
	switch (m_status)
	{
	case StatusOn:
		return (QObject::tr("unter Verkehr"));
	case StatusOff:
		return (QObject::tr("stillgelegt"));
	case StatusUnderConstruction:
		return (QObject::tr("Baustelle, nicht befahrbar"));
	default:
		return (QObject::tr("unbekannt"));
	}
}

QString CNode::GetTypeShortcut () const
{
	switch (m_typ)
	{
	default:
	case TypUnknown:
		return (QString());
	case TypAD:
		return ("AD");
	case TypAK:
		return ("AK");
	case TypAS:
		return ("AS");
	case TypAN:
		return ("AN");
	case TypBS:
		return ("BS");
	case TypBG:
		return ("BG");
	case TypTR:
		return ("TR");
	}
}

QString CNode::GetTypeText () const
{
	switch (m_typ)
	{
	default:
	case TypUnknown:
		return (QObject::tr("unbekannt"));
	case TypAD:
		return (QObject::tr("Autobahndreieck"));
	case TypAK:
		return (QObject::tr("Autobahnkreuz"));
	case TypAS:
		return (QObject::tr("Anschlussstelle"));
	case TypAN:
		return (QObject::tr("Anschluss"));
	case TypBS:
		return (QObject::tr("Behelfs-Anschluss"));
	case TypBG:
		return (QObject::tr("Bundesgrenze"));
	case TypTR:
		return (QObject::tr("Tank- und Raststtte"));
	}
}

QString CNode::GetUniqueString (void) const
{
QString strRet;
	strRet = QString ("A %1 %2").arg (m_iBAB).arg (m_strName);
	if (m_strNumber.find (QRegExp ("^[0-9]+[a-z]{0,1}$", false)) == 0)
		strRet += QString(" (%1)").arg (m_strNumber);
	return strRet;
}

QString CNode::AsHTML (void)
{
QString strRet, strWarning;
CLink *pLink=NULL;
CNode *pNode=NULL;

	strRet += "<h1>" + GetUniqueString () + "</h1>\n";

	strRet += QObject::tr("<table><tr><td>Type:</td><td>%1</td></tr><tr><td>Status:</td><td>%2</td></tr></table>").arg (GetTypeText()).arg(GetStatusText());
	strWarning = WarningText();
	if (!strWarning.isEmpty())
	{
		strRet += QObject::tr("<h2>Warnings</h2><i>Data file at line %1 ff.</i>").arg (m_uCSVLine);
		strRet += strWarning;
	}
	strRet += QObject::tr("<h2>Outgoing links</h2>\n");
	strRet += QObject::tr("<table><tr><th>BAB</th><th>Node</th><th>Name</th><th>Distance</th><th>Type</th><th>Status</th></tr>\n");
	for (pLink = m_listLinkOut.first(); pLink != NULL; pLink = m_listLinkOut.next())
	{
		pNode = pLink->m_pNodeTo;
		strRet += QObject::tr ("<tr><td>A %1</td><td>%2</td><td>%3</td><td>%4</td><td>%5</td><td>%6</td></tr>\n")
			.arg (pNode->m_iBAB)
			.arg (pNode->m_strNumber)
			.arg (pNode->m_strName)
			.arg (pLink->m_fKm, 0, 'f', 1)
			.arg (pNode->GetTypeText())
			.arg (pNode->GetStatusText());
	}
	strRet += "</table>\n";

	strRet += QObject::tr("<h2>Incoming links</h2>\n");
	strRet += QObject::tr("<table><tr><th>BAB</th><th>Node</th><th>Name</th><th>Distance</th><th>Type</th><th>Status</th></tr>\n");
	for (pLink = m_listLinkIn.first(); pLink != NULL; pLink = m_listLinkIn.next())
	{
		pNode = pLink->m_pNodeFrom;
		strRet += QObject::tr ("<tr><td>A %1</td><td>%2</td><td>%3</td><td>%4</td><td>%5</td><td>%6</td></tr>\n")
			.arg (pNode->m_iBAB)
			.arg (pNode->m_strNumber)
			.arg (pNode->m_strName)
			.arg (pLink->m_fKm, 0, 'f', 1)
			.arg (pNode->GetTypeText())
			.arg (pNode->GetStatusText());
	}
	strRet += "</table>\n";
	return strRet;
}

QString CNode::WarningText (void)
{
QString strRet;
CLink *pLink = NULL;

/*	if (m_strName.contains (QRegExp("\\bKreuz$")) && m_typ != TypAK)
		strRet += QObject::tr("<li>Name contains 'Kreuz' but type is not a highway cross (%1).</li>").arg (GetTypeText());
	else if (m_strName.contains (QRegExp("\\bDreieck$")) && m_typ != TypAD)
		strRet += QObject::tr("<li>Name contains 'Dreieck' but type is not a highway triangle (%1).</li>").arg (GetTypeText());
	else if (m_strName.contains (QRegExp("\bbergang\b")) && m_typ != TypAN)
		strRet += QObject::tr("<li>Name contains 'bergang' but type is not a transition (%1).</li>").arg (GetTypeText());
*/
	
	bool bCrossLink=false;
	for (pLink = m_listLinkOut.first(); pLink != NULL; pLink = m_listLinkOut.next())
	{
		if (pLink->m_pNodeFrom == pLink->m_pNodeTo)
			strRet += QObject::tr("<li>Node contains link to itself. (<i>Line: %1</i>).</li>").arg (pLink->m_uCSVLine);
		if (pLink->m_pNodeFrom->m_iBAB != pLink->m_pNodeTo->m_iBAB)
			bCrossLink=true;
	}

	if (m_listLinkIn.count() == 0)
		strRet += QObject::tr ("<li>Node has no incomming links</li>");
	if (m_listLinkOut.count() == 0)
		strRet += QObject::tr ("<li>Node has no outgoing links</li>");
	if (m_typ == TypBG && (m_listLinkIn.count() > 1 || m_listLinkOut.count() > 1))
		strRet += QObject::tr("<li>Node is a border crossing, but has more than one outgoing/incomming link.</li>");
	else if (m_typ == TypAS || m_typ == TypBS)
	{
		if (m_listLinkIn.count() > 2 || m_listLinkOut.count() > 2)
			strRet += QObject::tr("<li>Node is not a highway cross/triangle and has more than two outgoing/incomming links.</li>");
/*		if (m_listLinkIn.count() == 1 && m_listLinkOut.count() == 1)
		{
			if (m_listLinkIn.first()->m_pNodeFrom == m_listLinkOut.first()->m_pNodeTo)
				strRet += QObject::tr("<li>Node seems to be the end of the highway, but is not from type AN or BG.</li>");
		}
*/
	}
	else if (m_typ == TypAD && !bCrossLink && m_listLinkIn.count() < 3 && m_listLinkOut.count() < 3)
		strRet += QObject::tr("<li>Type is a highway triangle, but node hasn't links to other highways (Only connected to A %1).</li>").arg(m_iBAB);
	else if (m_typ == TypAK && !bCrossLink && m_listLinkIn.count() < 4 && m_listLinkOut.count() < 4)
		strRet += QObject::tr("<li>Type is a highway cross, but node hasn't links to other highways (Only connected to A %1).</li>").arg(m_iBAB);


/////	if ((m_typ == TypAK || m_typ == TypAD) && !bCrossLink)
/////		strRet += QObject::tr("<li>Type is a highway cross/triangle, but node hasn't links to other highways (Only connected to A %1).</li>").arg(m_iBAB);
	if (!(m_typ == TypAK || m_typ == TypAD) && bCrossLink)
		strRet += QObject::tr("<li>Type is not a highway cross/triangle, but node has links to other highways.</li>");

	if (!strRet.isEmpty()) strRet = "<ul>"+strRet+"</ul>";
	return strRet;
}

QPixmap CNode::GetPixmap16 (void) const
{
const char* autobahn[] = { 
"16 16 8 1",
"f c #0033cc",
". c #0033ff",
"e c #3366ff",
"b c #99cccc",
"c c #99ccff",
"d c #cccccc",
"a c #ccccff",
"# c #ffffff",
"................",
".......##.......",
".......##.......",
"......####......",
"......####......",
".....a#bc#d.....",
".##############.",
"...e.aaeeaa.e...",
"....e##..##e....",
"....###..###....",
"....###..###....",
"...####..####...",
"...####..####...",
"...####..####...",
"..a####..####a..",
"................"};

const char* bau[] = { 
"16 16 10 1",
". c None",
"d c #000000",
"c c #232323",
"h c #3f3f00",
"f c #474747",
"e c #8f8f8f",
"g c #d3d3f7",
"b c #dbdbdb",
"# c #ff0000",
"a c #ffffff",
"................",
".......##.......",
"......####......",
"......#aa#......",
".....##aa##.....",
".....#aaaa#.....",
"....##abca##....",
"....#cddaaa#....",
"...##dddaaa##...",
"...#aedcaaaa#...",
"..##addafdda##..",
"..#acagadddca#..",
".##aaaahcdcca##.",
".##############.",
".##############.",
"................"};

const char* zoll[] = { 
"16 16 32 1",
". c None",
"s c #000000",
"y c #303030",
"x c #515151",
"z c #5f5f5f",
"C c #636363",
"w c #6a6a6a",
"r c #747474",
"n c #7a7a7a",
"l c #838383",
"u c #8b8b8b",
"D c #8d6e6e",
"v c #939393",
"p c #949494",
"q c #999999",
"t c #9d9d9d",
"A c #a3a3a3",
"b c #a7a7a7",
"j c #a9a9a9",
"m c #afafaf",
"k c #b0b0b0",
"i c #bdbdbd",
"d c #c0c0c0",
"o c #c5c5c5",
"B c #c9c9c9",
"c c #d4d4d4",
"f c #e3e3e3",
"e c #efefef",
"g c #fafafa",
"# c #ff0000",
"h c #ffe0e0",
"a c #ffffff",
"................",
"....########....",
"...##########...",
"..###aaaaaa###..",
".###bcdefgfh##..",
".##aijklmebaa##.",
".##anopqnorca##.",
".##assssssssa##.",
".##assssssssa##.",
".##atrmunvlwa##.",
".##axidyzxxva##.",
".###AqtlBpCD##..",
"..###aaaaaa###..",
"...##########...",
"....########....",
"................"};

const char* verbot[] = { 
"16 16 3 1",
". c None",
"# c #ff0000",
"a c #ffffff",
"................",
"....########....",
"...##########...",
"..###aaaaaa###..",
".###aaaaaaaa##..",
".##aaaaaaaaaa##.",
".##aaaaaaaaaa##.",
".##aaaaaaaaaa##.",
".##aaaaaaaaaa##.",
".##aaaaaaaaaa##.",
".##aaaaaaaaaa##.",
".###aaaaaaaa##..",
"..###aaaaaa###..",
"...##########...",
"....########....",
"................"};

const char* rast[] = { 
"16 16 8 1",
". c #0033ff",
"c c #171719",
"f c #3d3d47",
"a c #565662",
"d c #7d7d8b",
"b c #95959d",
"e c #bebec6",
"# c #fbfbfc",
"................",
"................",
"................",
"..############..",
"..#aa#####ba##..",
"..#bcd###aaae#..",
"..##bca##ccb##..",
"..###bcaad####..",
"..#####fa#####..",
"..####adaf####..",
"..###ca##aa###..",
"..##ca####ac##..",
"..#aa######fd#..",
"..############..",
"................",
"................"};

const char* kreuz[] = { 
"16 16 7 1",
". c #0133ff",
"c c #2e57fd",
"b c #5071fe",
"# c #7992fd",
"d c #94a9fd",
"a c #b9c7fe",
"e c #e2e7fe",
"................",
"................",
"................",
"..#ab......cdd..",
"ceddeedd#deee#eb",
".dee#eebcaeddea.",
"..ededdeeddeaec.",
"..dccee#aeab.d..",
"..dcdeded#eeca..",
".ceea#edaeddeeb.",
"cee#aedcc#ee#eeb",
".bdeedbbbbdeed#c",
"..cb........bb..",
"................",
"................",
"................"};

const char* ausfahrt[] = { 
"16 16 3 1",
". c #0033ff",
"# c #3366ff",
"a c #f9fbfe",
"................",
"................",
"....#a.a...aaaaa",
"....aa.a....aaa.",
"....aa.aa..aaaa.",
"....aa.aaaaa.a..",
"...aaa.aaaa..a..",
"...aaa.aaa......",
"...aaa.aaa......",
"..aaaa.aaa......",
"..aaaa.aaaa.....",
"..aaaa.aaaa.....",
".aaaaa.aaaa.....",
"................",
"................",
"................"};

	if (m_status == StatusOff)
		return (QPixmap (verbot));
	else if (m_status == StatusUnderConstruction)
		return (QPixmap (bau));
	switch (m_typ)
	{
	default:
	case TypUnknown:
		return (QPixmap ());
	case TypAD:
	case TypAK:
		return (QPixmap (kreuz));
	case TypAS:
		return (QPixmap (ausfahrt));
	case TypAN:
		return (QPixmap (autobahn));
	case TypBS:
		return (QPixmap (ausfahrt));
	case TypBG:
		return (QPixmap (zoll));
	case TypTR:
		return (QPixmap (rast));

	}
}

bool CNode::IsOneWay ()
{
	if (m_listLinkIn.count() != 1 || m_listLinkOut.count() != 1) return false;
	if (m_listLinkIn.getFirst()->m_pNodeFrom == m_listLinkOut.getFirst()->m_pNodeTo) return false;

	return true;
}

CNode *CNode::FindNextSubNodeToDisplay (void)
{
CLink *pLink = NULL;
int iCount=0;
CNode *pNodeRet=NULL;

	for (pLink = m_listLinkOut.first(); pLink; pLink = m_listLinkOut.next())
	{
		if (pLink->m_pNodeTo->m_pCanvasNode != NULL) continue;
		if (pLink->m_pNodeTo->m_iBAB != m_iBAB) return (pLink->m_pNodeTo);
	}
	if (m_typ == TypAD || m_typ == TypAK)
	{
		for (pLink = m_listLinkOut.first(); pLink; pLink = m_listLinkOut.next())
		{
			if (pLink->m_pNodeTo->m_iBAB != m_iBAB) continue;
			iCount++;
			if (!pNodeRet 
				&& pLink->m_pNodeTo->m_listLinkOut.count() == 1 
				&& pLink->m_pNodeTo->m_listLinkOut.getFirst()->m_pNodeTo == this
				&& !pLink->m_pNodeTo->m_pCanvasNode)
			{
				pNodeRet = pLink->m_pNodeTo;
			}
		}
		if (iCount == 3 && pNodeRet) return pNodeRet;
	}
	return NULL;
}

CNode *CNode::FindNextLeftNodeToDisplay (void)
{
CLink *pLink = NULL;
CNode *pNode=NULL, *pNextNode=NULL;
QPtrList<CNode> list;

	pNextNode = this;
	while (pNextNode)
	{
		if (list.contains (pNextNode)) break;
		list.append (pNextNode);
		pNode = pNextNode;
		pNextNode = NULL;
		for (pLink = pNode->m_listLinkOut.first(); !pNextNode && pLink; pLink = pNode->m_listLinkOut.next())
		{
			if (pLink->m_pNodeTo->m_iBAB != m_iBAB) continue;
			if (pLink->m_pNodeTo == this) continue;
			if (!pLink->m_pNodeTo->IsOneWay()) continue;

			if (pLink->m_pNodeTo->m_pCanvasNode != NULL)
			{
				pNextNode = pLink->m_pNodeTo;
				break;
			}
			else
				return pLink->m_pNodeTo;
		}
	}
	return NULL;
}

CNode *CNode::FindNextRightNodeToDisplay (void)
{
CLink *pLink = NULL;
CNode *pNode=NULL, *pNextNode=NULL;
QPtrList<CNode> list;

	pNode = this;
	pNextNode = this;
	while (pNextNode)
	{
		if (list.contains (pNextNode)) break;
		list.append (pNextNode);
		pNode = pNextNode;
		pNextNode = NULL;
		for (pLink = pNode->m_listLinkIn.first(); !pNextNode && pLink; pLink = pNode->m_listLinkIn.next())
		{
			if (pLink->m_pNodeFrom->m_iBAB != m_iBAB) continue;
			if (!pLink->m_pNodeFrom->IsOneWay()) continue;

			if (pLink->m_pNodeFrom->m_pCanvasNode != NULL)
			{
				pNextNode = pLink->m_pNodeFrom;
				break;
			}
			else
				return pLink->m_pNodeFrom;
		}
	}
	return NULL;
}

void CNode::debug (void) const
{
	qDebug (QString ("Knoten BAB %1 Nr. %2: %3 (Typ: %4 Status: %5)").arg (m_iBAB).arg(m_strNumber).arg(m_strName).arg(GetTypeShortcut()).arg(GetStatusText()));
}

bool operator == (CNode& n1, CNode& n2)
{
	return (n1.m_iBAB == n2.m_iBAB 
		&& n1.m_strName == n2.m_strName 
		&& n1.m_strNumber == n2.m_strNumber
		&& n1.m_typ == n2.m_typ);
}

CLink::CLink ()
{
	m_pNodeFrom = 0;
	m_pNodeTo = 0;
	m_fKm = 0.0;
	m_uCSVLine = 0;
}

CLink::CLink (CNode *pNodeFrom, CNode *pNodeTo, const float fKm, const unsigned uCSVLine)
{
	m_pNodeFrom = pNodeFrom;
	m_pNodeTo = pNodeTo;
	m_fKm = fKm;
	m_uCSVLine = uCSVLine;

	m_pNodeFrom->m_listLinkOut.append (this);
	m_pNodeTo->m_listLinkIn.append(this);

	m_pCanvasLink = NULL;
}

CLink::~CLink ()
{
}

CNetwork::CNetwork ()
{
	m_listLink.setAutoDelete (true);
	m_listNode.setAutoDelete (true);
	m_pCalculatedFrom = NULL;
}

CNetwork::~CNetwork ()
{
}

void CNetwork::CreateSampleNetwork (void)
{
CNode *pNode[9];
int i;
	for (i=0; i<9; i++)
	{
		pNode[i] = new CNode ();
		pNode[i]->m_iBAB = i+1;
		m_listNode.append (pNode[i]);
	}

	m_listLink.append (new CLink (pNode[0], pNode[1], 2));
	m_listLink.append (new CLink (pNode[1], pNode[0], 2));

	m_listLink.append (new CLink (pNode[0], pNode[6], 15));
	m_listLink.append (new CLink (pNode[6], pNode[0], 15));

	m_listLink.append (new CLink (pNode[0], pNode[5], 9));
	m_listLink.append (new CLink (pNode[5], pNode[0], 9));
	m_listLink.append (new CLink (pNode[1], pNode[6], 6));
	m_listLink.append (new CLink (pNode[6], pNode[1], 6));
	m_listLink.append (new CLink (pNode[1], pNode[2], 4));
	m_listLink.append (new CLink (pNode[2], pNode[1], 4));
	m_listLink.append (new CLink (pNode[2], pNode[3], 2));
	m_listLink.append (new CLink (pNode[3], pNode[2], 2));
	m_listLink.append (new CLink (pNode[2], pNode[8], 15));
	m_listLink.append (new CLink (pNode[8], pNode[2], 15));
	m_listLink.append (new CLink (pNode[8], pNode[3], 1));
	m_listLink.append (new CLink (pNode[3], pNode[8], 1));
	m_listLink.append (new CLink (pNode[6], pNode[8], 2));
	m_listLink.append (new CLink (pNode[8], pNode[6], 2));
	m_listLink.append (new CLink (pNode[6], pNode[7], 15));
	m_listLink.append (new CLink (pNode[7], pNode[6], 15));
	m_listLink.append (new CLink (pNode[8], pNode[7], 4));
	m_listLink.append (new CLink (pNode[7], pNode[8], 4));
	m_listLink.append (new CLink (pNode[3], pNode[4], 1));
	m_listLink.append (new CLink (pNode[4], pNode[3], 1));
	m_listLink.append (new CLink (pNode[7], pNode[4], 3));
	m_listLink.append (new CLink (pNode[4], pNode[7], 3));
	m_listLink.append (new CLink (pNode[4], pNode[5], 6));
	m_listLink.append (new CLink (pNode[5], pNode[4], 6));
	m_listLink.append (new CLink (pNode[5], pNode[7], 11));
	m_listLink.append (new CLink (pNode[7], pNode[5], 11));
	CalculateNearestDistances (pNode[0]);

}

bool CNetwork::Load (const QString strFile, QWidget *pParent, const bool bForce)
{
QString strLine, strStatus, str, strOldStatus;
QStringList strl;
QFile file (strFile);
CNode node, *pNodeFrom=NULL, *pNodeTo=NULL;
CLink *pLink=NULL;
float fKm = 0.0;
unsigned uLine=0;

	if (strFile.isEmpty()) return false;
	if (strFile == m_strDBFile && !bForce) return true;

	if (!file.open (IO_ReadOnly))
	{
		QMessageBox::critical (pParent, QObject::tr("Error"), QObject::tr("Could not open data file %1").arg (strFile));
		return false;
	}

QTextStream stream (&file);
QProgressDialog dlg (QObject::tr("Reading database..."), QObject::tr ("&Cancel"), file.size(), pParent);
	strLine = stream.readLine ();		// first line with header
	uLine = 1;

	m_strDBFile = QString ();	
	m_listLink.clear();
	m_listNode.clear();
	while (!stream.atEnd ())
	{
		if (uLine % 10 == 0) dlg.setProgress (file.at());
		qApp->processEvents();
		if (dlg.wasCanceled())
		{
			m_listLink.clear();
			m_listNode.clear();
			return false;	
		}

		strLine = stream.readLine ();
		uLine++;
		strl = QStringList::split (";", strLine, true);
		strOldStatus = strStatus;
		strStatus = strl[8].stripWhiteSpace();

		node.m_iBAB = strl[0].toInt();
		node.m_strNumber = strl[1].stripWhiteSpace();
		node.m_typ = CNode::GetTyp (strl[2]);
		node.m_strName = strl[3].stripWhiteSpace();
		node.m_status = CNode::GetStatus (strStatus[0].latin1());
		node.m_uCSVLine = uLine;

		//if (pNodeTo && strOldStatus != "rr" && (strStatus != "rr" || !(node == *pNodeTo)))
		if (pNodeTo && strOldStatus != "rr" && strStatus != "rr" && pNodeTo != pNodeFrom)
		{
			pLink = new CLink(pNodeTo, pNodeFrom, fKm, uLine-1);
			m_listLink.append (pLink);
		}

		pNodeFrom = AddNode (node);

		node.m_iBAB = pNodeFrom->m_iBAB;
		node.m_strNumber = strl[4].stripWhiteSpace();
		node.m_typ = CNode::GetTyp (strl[5]);
		node.m_strName = strl[6].stripWhiteSpace();
		node.m_status = CNode::GetStatus (strStatus[1].latin1());
		node.m_uCSVLine = uLine;
		pNodeTo = AddNode (node);

		str = strl[7];
		str.replace (",", ".");
		fKm = str.toFloat();
		pLink = new CLink(pNodeFrom, pNodeTo, fKm, uLine);
		m_listLink.append (pLink);
	}
	file.close();

	CreateCrossLinks ();
	m_strDBFile = strFile;
	return true;
}

CNode *CNetwork::AddNode (CNode &node)
{
CNode *pNode;
	m_listNode.first();
	while (m_listNode.current())
	{
		pNode = m_listNode.current();
		if ((*pNode) == node) {/*qDebug("Found:"); pNode->debug();*/ return pNode;}
		m_listNode.next();
	}
	pNode = new CNode (node);
	m_listNode.append (pNode);
	//pNode->debug();
	return pNode;
}

void CNetwork::CreateCrossLinks (void)
{
QPtrListIterator<CNode> it1(m_listNode), it2(m_listNode);
CNode *pNode1=NULL, *pNode2=NULL;
CLink *pLink=NULL;
	
	while ((pNode1 = it1.current()) != NULL)
	{
		if (!pNode1->IsCross()) 
		{
			++it1;
			continue;
		}
		it2 = it1;
		++it2;
		while ((pNode2 = it2.current()) != NULL)
		{
			++it2;
			if (pNode2->m_typ != pNode1->m_typ) continue;
			if (pNode2->m_strName != pNode1->m_strName) continue;
			pLink =  new CLink (pNode1, pNode2, 0.0);
			m_listLink.append (pLink);
			pLink = new CLink (pNode2, pNode1, 0.0);
			m_listLink.append (pLink);
		}
		++it1;
	}
}


void CNetwork::CalculateNearestDistances (CNode *pNodeFrom)
{
QPtrList<CNode> listNodeTodo;
CNode *pNearestNode=NULL, *pNode=NULL;
CLink *pLink=NULL;
	
	if (m_pCalculatedFrom == pNodeFrom) return;

	// initialize
	m_listNode.first ();
	while ((pNode = m_listNode.current()) != NULL)
	{
		pNode->m_pNearestLink = NULL;
		pNode->m_fNearestKm = 9999999;
		m_listNode.next();
	}
	pNodeFrom->m_fNearestKm = 0.0;
	listNodeTodo.append (pNodeFrom);

	while (!listNodeTodo.isEmpty())
	{
		// find nearest node
		pNearestNode = NULL;
		listNodeTodo.first ();
		while ((pNode = listNodeTodo.current()) != NULL)
		{
			if (pNode == pNodeFrom || pNearestNode == NULL || pNode->m_fNearestKm < pNearestNode->m_fNearestKm)
				pNearestNode = pNode;
			listNodeTodo.next();
		}

		//
		pNearestNode->m_listLinkOut.first ();
		while ((pLink = pNearestNode->m_listLinkOut.current()) != NULL)
		{
			if (pLink->m_pNodeTo->m_pNearestLink == NULL && !listNodeTodo.containsRef (pLink->m_pNodeTo) && pLink->m_pNodeTo != pNodeFrom)
				listNodeTodo.append (pLink->m_pNodeTo);
			if ((pLink->m_pNodeTo != pNodeFrom && pLink->m_pNodeTo->m_pNearestLink == NULL) || pLink->m_pNodeTo->m_fNearestKm > pNearestNode->m_fNearestKm + pLink->m_fKm)
			{
				pLink->m_pNodeTo->m_pNearestLink = pLink;
				pLink->m_pNodeTo->m_fNearestKm = pNearestNode->m_fNearestKm + pLink->m_fKm;
			}
			pNearestNode->m_listLinkOut.next ();
		}
		listNodeTodo.removeRef (pNearestNode);
	}
	m_pCalculatedFrom = pNodeFrom;
}

void CNetwork::CalculateRoute (CNode *pNodeFrom, CNode *pNodeTo, QPtrList<CLink> &listLink)
{
CNode *pNode = pNodeTo;
	CalculateNearestDistances (pNodeFrom);

	listLink.clear();
	listLink.setAutoDelete (false);

	while (pNode != NULL && pNode->m_pNearestLink != NULL)
	{
		listLink.prepend (pNode->m_pNearestLink);
		pNode = pNode->m_pNearestLink->m_pNodeFrom;
	}
}

void CNetwork::FindNode (const QString strSearch, QPtrList<CNode> &listNode, const bool bAllowCrosses)
{
QStringList strl = QStringList::split (" ", strSearch, false);
QString str, strNumber;
int iBAB = 0;
CNode *pNode = NULL;
bool bFound=true;

	// look for BAB-number
	QStringList::iterator it;
	for (it = strl.begin(); it != strl.end(); ++it )
	{
		str = (*it);
		if (iBAB == 0 && (str.upper() == "A" || str.upper() == "BAB"))
		{
			it = strl.remove (it);
			if (it == strl.end()) break;
			str = (*it);
			iBAB = str.toInt ();
			it = strl.remove (it);
		}
		else if (str.length() < 3) it = strl.remove (it);
		if (it == strl.end()) break;

		if (iBAB == 0 && str.find (QRegExp ("^(A|BAB)[0-9]{1,3}$", false)) != -1)
		{
			str.remove (QRegExp ("^(A|BAB)", false));
			iBAB = str.toInt ();
			it = strl.remove (it);
		}
		else if (strNumber.isEmpty() && str.find (QRegExp ("^\\x0028[0-9]+[a-z]{0,1}\\x0029$", false)) != -1)
		{
			strNumber = str;
			strNumber.remove ("(").remove (")");
			it = strl.remove (it);
		}
	}

	listNode.clear ();
	listNode.setAutoDelete (false);

	for (pNode = m_listNode.first(); pNode != NULL; pNode = m_listNode.next())
	{
		if (!bAllowCrosses && 
			(pNode->m_typ == CNode::TypAD 
			|| pNode->m_typ == CNode::TypAK 
			|| pNode->m_typ == CNode::TypTR)) continue;
		if (iBAB != 0 && pNode->m_iBAB != iBAB) continue;
		if (!strNumber.isEmpty() && pNode->m_strNumber.lower() != strNumber.lower()) continue;
		bFound = true;
		for (it = strl.begin(); it != strl.end() && bFound; ++it )
		{
			str = (*it);
			if (!pNode->m_strName.contains (str, false)) bFound = false;
		}
		if (bFound)
			listNode.append (pNode);
	}
}

CNode* CNetwork::FindNextNodeToDisplay (const int iBAB, CNode *pOldNode)
{
CNode *pNode=NULL, *pNextNode=NULL;
CLink *pLink=NULL;
QPtrList<CNode> list;
bool bOK=true;

	pNextNode = pOldNode;
	while (pNextNode)
	{
		if (list.contains (pNextNode)) break;
		list.append (pNextNode);
		pNode = pNextNode;
		pNextNode = NULL;
		for (pLink = pNode->m_listLinkOut.first(); !pNextNode && pLink; pLink = pNode->m_listLinkOut.next())
		{
			if (pLink->m_pNodeTo->m_iBAB != iBAB) continue;
			if (pLink->m_pNodeTo->IsOneWay())
			{
				pNextNode = pLink->m_pNodeTo;
				break;
			}
			if (pLink->m_pNodeTo->m_pCanvasNode != NULL) continue;
			return (pLink->m_pNodeTo);
		}
	}

	for (pNode = m_listNode.first(); pNode != NULL; pNode = m_listNode.next())
	{
		if (pNode->m_iBAB != iBAB) continue;
		if (pNode->m_pCanvasNode == NULL) return (pNode);
	}
	return NULL;
}

QString CNetwork::Check (void)
{
QString strRet, str;
CNode *pNode=NULL;
CLink *pLink=NULL;
bool bHasError = false;
QPtrList<CNode> listNode;

	strRet = QObject::tr("<h1>Database check</h1>\n");
	if (!m_strDBFile.isEmpty())
		strRet += QObject::tr("<p><i>File: %1</i></p>\n").arg(m_strDBFile);
	for (pNode = m_listNode.first(); pNode != NULL; pNode = m_listNode.next())
	{
		str = pNode->WarningText();
		if (str.isEmpty()) continue;
		strRet +="<h3>" + pNode->GetUniqueString() + "</h3>\n" + QObject::tr("<i>Line %1</i><br>\n").arg(pNode->m_uCSVLine) + str;
		bHasError = true;
	}
	if (m_strDBFile.isEmpty())
		strRet += QObject::tr("No database loaded.");
	else if (!bHasError) 
		strRet += QObject::tr("No Errors found.");

	if (m_pCalculatedFrom == NULL)
	{
		FindNode ("ulm-west", listNode);
		pNode = listNode.first();
		CalculateNearestDistances (pNode);
	}
	str=QString();
	for (pNode = m_listNode.first(); pNode != NULL; pNode = m_listNode.next())
	{
		if (pNode == m_pCalculatedFrom) continue;
		if (pNode->m_pNearestLink == NULL)
			str += "<li>"+pNode->GetUniqueString()+"</li>\n";
	}
	if (!str.isEmpty())
	{
		strRet += QObject::tr("<h2>Mesh-Check</h2>");
		strRet += QObject::tr("<p>You can not reach the following nodes through the highway network, starting from <b>%1</b>:</p><ul>%2</ul>").arg(m_pCalculatedFrom->GetUniqueString()).arg(str);
	}

	return (strRet);
};
